// /*******************************************************************************
//  * 创建: 2022年11月27日
//  * 作者: 水煮肉片饭(27185709@qq.com)
//  * 描述: 动作残影
//  *      给节点添加残影，残影会跟随节点移动，并跟随节点播放动画。
//  *      使用举例：动作类游戏中，主角释放连击必杀
// *******************************************************************************/

import { CCObject,_decorator, Component, CCInteger, Color, color, Animation, Node, UIOpacity, UIRenderer, instantiate, Sprite, UITransform, isValid, Mat4, spriteAssembler, SkeletalAnimation } from 'cc';
import { DEV, EDITOR } from 'cc/env';
const { ccclass, property, executeInEditMode, playOnFocus, menu } = _decorator;

class ShadowData {
    x: number = 0;
    y: number = 0;
    angle: number = 0;
    scaleX: number = 1;
    scaleY: number = 1;
    scale: number = 1;          //向尾部递减的缩放系数
    actionName: string = '';
}

@ccclass('ActionShadow')
@executeInEditMode@playOnFocus@menu('Comp/ActionShadow')
export class ActionShadow extends Component {
    @property
    private _active: boolean = true;
    @property({ displayName: DEV && '是否激活', tooltip: DEV && '设置残影可见性\n激活会重置残影位置' })
    get active() { return this._active; }
    set active(value: boolean) {
        this._active = value;
        if (value) {
            if (this.node.active && this.shadowNode) {
                this.shadowNode.active = true;
                this.shadowData.length && this.updateShadowData();
            }
        } else {
            this.shadowNode.active = false;
        }
    }

    @property
    _shadowNum:number = 10;
    @property({ type: CCInteger, displayName: DEV && '残影数量' })
    private get shadowNum(): number { return this._shadowNum; }
    private set shadowNum(value: number) {
        this._shadowNum = Math.max(value, 0);
        this.updateShadowNum();
        this.updateDeltTime();
        this.updateShadowData();
        this.updateColor();
    }

    @property
    private _deltTime: number = 4;
    @property({ type: CCInteger, displayName: DEV && '延迟帧数' })
    private get deltTime(): number { return this._deltTime; }
    private set deltTime(value: number) {
        this._deltTime = Math.max(value, 1);
        this.updateDeltTime();
        this.updateShadowData();
    }

    @property
    private _shadowColor: Color = color(255, 255, 255);
    @property({ displayName: DEV && '残影颜色' })
    private get shadowColor(): Color { return this._shadowColor; }
    private set shadowColor(value: Color) {
        this._shadowColor = value;
        this.updateColor();
    }

    @property
    private _shadowScale: number = 0.1;
    @property({ min: 0, max: 1, step: 0.1, slide: true, displayName: DEV && '尾部缩放系数' })
    private get shadowScale(): number { return this._shadowScale; }
    private set shadowScale(value: number) {
        this._shadowScale = value;
        this.updateShadowData();
    }

    @property
    private _opacity: number = 100;
    @property({ type: CCInteger, min: 0, max: 255, slide: true, displayName: DEV && '透明度' })
    private get opacity(): number { return this._opacity; }
    private set opacity(value: number) {
        this._opacity = value;
        this.updateOpacity();
    }

    private nodeOpacity: number = 255;
    private model: Animation | null = null;
    private shadowNode: Node | null = null;
    private shadowData: ShadowData[] = [];

    protected start() {
        let shadowNodeName = this.node.name + '<ActionShadow>';
        this.shadowNode = this.node.parent.getChildByName(shadowNodeName);
        if (!this.shadowNode) {
            console.log("shadowNode new");
            this.shadowNode = new Node(shadowNodeName);
            this.shadowNode.addComponent(UIOpacity);
            this.shadowNode.addComponent(Sprite);
            this.shadowNode.setParent(this.node.parent);
            this.shadowNode.setSiblingIndex(this.node.getSiblingIndex());
        }
        if (EDITOR) {
            this.shadowNode._objFlags = 0;
            // console.log("Object['Flags']:",CCObject['Flags'])
            // console.log("Object['Flags'].LockedInEditor:",CCObject['Flags'].LockedInEditor)
            this.shadowNode['_objFlags'] |= CCObject['Flags'].HideInHierarchy;
            this.shadowNode['_objFlags'] |= CCObject['Flags'].LockedInEditor;
        }
        this.nodeOpacity = this.node.getComponent(UIOpacity)?.opacity;
        this.model = this.node.getComponent(Animation);
        // this.model && (this.model.currentClip = this.model.defaultClip);
        
        this.node.updateWorldTransform();
        this.updateShadowNum();
        this.updateDeltTime();
        this.updateShadowData();
        this.updateColor();
        this.updateOpacity();
    }

    protected onEnable() {
        if (this.active && this.shadowNode) {
            this.shadowNode.active = true;
            this.shadowData.length && this.updateShadowData();
        }
    }

    protected onDisable() {
        this.shadowNode.active = false;
    }

    protected update() {
        if (this.nodeOpacity !== this.node.getComponent(UIOpacity)?.opacity) {
            this.nodeOpacity = this.node.getComponent(UIOpacity)?.opacity;
            this.updateOpacity();
        }
        for (let i = this.shadowNum * this.deltTime; i > 0; --i) {
            let cur = this.shadowData[i];
            let prev = this.shadowData[i - 1];
            cur.x = prev.x;
            cur.y = prev.y;
            cur.scaleX = prev.scaleX;
            cur.scaleY = prev.scaleY;
            cur.angle = prev.angle;
            cur.actionName = prev.actionName;
        }

        let data = this.shadowData[0];
        let matrix:Mat4 = this.node.worldMatrix;
        data.x = matrix.m12;
        data.y = matrix.m13;
        data.scaleX = this.node.scale.x;
        data.scaleY = this.node.scale.y;
        data.angle = this.node.angle;
        if (this.model !== null) {
            data.actionName = this.model.defaultClip.name;
        }
        this.shadowNode.updateWorldTransform();
        matrix = this.shadowNode.worldMatrix;

        for (let i = this.shadowNum - 1; i >= 0; --i) {
            let node = this.shadowNode.children[i];
            data = this.shadowData[this.deltTime * (i + 1)];
            let x = data.x - matrix.m12;
            let y = data.y - matrix.m13;
            node.setPosition(x,y);
            node.setScale(data.scaleX * data.scale,data.scaleY * data.scale);
            node.angle = data.angle;
            let model = node.getComponent(Animation);
            model !== null && model.play(data.actionName);
        }
    }

    private updateShadowNum() {
        this.shadowNode.removeAllChildren();
        this.shadowNode.destroyAllChildren();
        for (let i = 0, len = this.shadowNum; i < len; ++i) {
            let node = instantiate(this.node);
            node.name = this.node.name + i;
            let cmps = node['_components'];
            for (let j = cmps.length - 1; j >= 0; --j) {
                if (cmps[j] instanceof UIRenderer) continue;
                if (cmps[j] instanceof Animation) continue;
                // if (cmps[j] instanceof SkeletalAnimation) continue;
                cmps[j].destroy();
            }
            node.setParent(this.shadowNode);
        }
    }

    private updateDeltTime() {
        this.shadowData = [];
        for (let i = this.shadowNum * this.deltTime; i >= 0; --i) {
            this.shadowData[i] = new ShadowData();
        }
    }

    private updateShadowData() {
        let scaleDelt = (1 - this.shadowScale) / (this.shadowNum * this.deltTime);
        let matrix:Mat4 = this.node.worldMatrix;
        for (let i = this.shadowNum * this.deltTime; i >= 0; --i) {
            let data = this.shadowData[i];
            data.x = matrix.m12;
            data.y = matrix.m13;
            data.scaleX = this.node.scale.x;
            data.scaleY = this.node.scale.y;
            data.scale = 1 - i * scaleDelt;
            data.angle = this.node.angle;
            if (this.model !== null) {
                data.actionName = this.model.defaultClip.name;
                this.model.play(data.actionName);
            }
        }
    }
    
    private updateColor() {
        for (let index = 0; index < this.shadowNode.children.length; index++) {
            const element = this.shadowNode.children[index];
            element.getComponentInChildren(Sprite).color = this.shadowColor;
        }
    }

    private updateOpacity() {
        let component = this.shadowNode.getComponent(UIOpacity);
        if(component){
            component.opacity = this.opacity;
        }
    }

    protected onDestroy() {
        if (isValid(this.shadowNode)) {
            this.shadowNode.removeFromParent();
            this.shadowNode.destroy();
        };
    }
}